# -*- coding: utf-8 -*-

import os

from stone import config
from stone import backends
from stone import generators
from stone import renderers
from stone.page import Page
from stone.site import Site


def generate_site(args):
    """Generate the described sites

    Load the config for each site and the require generators, renderers and
    backends for each.
    """

    site_configs = config.read(args.site_root)
    sites = (Site(args.site_root, data=config.data) for config in site_configs)

    for site in sites:
        # If not site type is selected we will presume it is a single page
        site_type = site.get('type', 'single')
        generator_name = generators.get_module_name(site_type)
        generator = generators.load_generator(generator_name)

        # Renderers are chained together in a decorator pattern. Once the chain
        # is created we can treat it as a single renderer as a page will go
        # through the specified rendering process.
        _renderer = renderers.load_renderers(
            renderers=site.get('renderers', []),
            templates=site.get('templates', []))

        _backends = backends.load_backends(
            backends=site.get('backends', []), root=args.site_root)

        for page in generator.generate(_renderer, site):
            for backend in _backends:
                backend.commit(page)


def new_page(args):
    """Add new page to the site"""
    site_configs = config.read(args.site_root)
    sites = (Site(args.site_root, data=config.data) for config in site_configs)
    if not hasattr(args, 'site'):
        print('What site would you like to add a new page to?')
        count = 0
        for site in sites:
            print("%i - %s" % (count, site['site']))
            count += 1

        choice = input()
        if not int(choice) < len(sites):
            print("[ERROR] %s is not a valid selection" % choice)

    site = sites[int(choice)]
    if site:
        create_add_page(
            site, args.source, args.target, data={'page_type': args.page_type})
        config.write(args.site_root, sites)


def create_add_page(site, source, target, data=None, content=None):
    """Create a Page() and file on disk"""
    init_content = '# Hello, World!'
    if content is None and not isinstance(content, str):
        content = init_content

    try:
        os.mkdir(os.path.join('{}/{}'.format(site.root, site['source'])))
    except FileExistsError:
        pass

    file_path = '{}/{}/{}'.format(site.root, site['source'], source)
    file = open(file_path, 'w')
    file.write(content)
    file.close()

    if target is None:
        # Attempt to sanitize the filename from source for our output.html
        from unidecode import unidecode
        target = unidecode(source)
        target = target.lower().replace(r' ', '-')
        target = '{}.html'.format(target.split('.')[0])

    site.add_page(Page(site, source, target, data))


def init_site(args):
    """Creata a new site from template"""
    index_content = '''
{% for post in posts %}
    <ul>
        <li><a href="{{ post.href }}">{{ post.title }}</a></li>
    </ul>
{% endfor %}
'''
    init_content = 'title: Hello, World!\n\n# Hello, World!'

    site = Site(args.site_root, {
        'site': args.site_name,
        'source': 'source',
        'target': 'target'
    })
    if args.type == 'blog':
        create_add_page(
            site,
            'index.md',
            'index.html',
            content=index_content,
            data={'page_type': 'index'})
        create_add_page(
            site,
            'example.md',
            'example.html',
            content=init_content,
            data={'page_type': 'post'})
    else:
        create_add_page(site, 'index.md', 'index.html', content=init_content)

    config.write(args.site_root, [site])



"""Abstract base class for SiteGenerators"""

from abc import (
    ABC,
    abstractmethod,
)


class SiteGenerator(ABC):
    pass

    @abstractmethod
    def generate(self, renderer, site):
        pass



"""Page

Stone's representation of a page
"""
from collections import UserDict
from json import JSONEncoder
import os
from typing import Any, Dict


class PageEncoder(JSONEncoder):
    """JSON encoder for Page"""

    def default(self, o):
        if isinstance(o, Page):
            return o.to_entry()

        # When not a page, call the JSONEncoder. It will call the correct fail.
        return JSONEncoder.default(self, o)


class Page(UserDict):  # pylint: disable=too-many-ancestors
    """Representation of a Page"""

    _site = None
    data: Dict[str, str] = {}

    def __init__(self, site, source: str, target: str,
                 data: Dict = None) -> None:
        super().__init__()
        self.data = {}
        self._site = site

        def get(key, dic):
            """Get a dictionary key if it exists"""
            return dic[key] if key in dic else None

        try:
            for k in data.keys():
                self.data[k] = get(k, data)
        except AttributeError:
            pass

        self.data["source"] = source
        self.data["target"] = target
        self.data["source_path"] = os.path.abspath(
            os.path.join(self._site.root, site['source'], source))
        self.data["target"] = target
        self.data["target_path"] = os.path.abspath(
            os.path.join(self._site.root, site['target'], target))
        try:
            self.data["href"] = target.split('/')[1]
        except IndexError:
            self.data["href"] = target
        self.data['content'] = open(self.data['source_path'], "r").read()

    def __contains__(self, key):
        return key in self.data

    def __delitem__(self, key):
        del self.data[key]

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        self.data.__iter__()

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        class_name = type(self).__name__
        return "{}({}, {})".format(class_name, self['source'], self['target'])

    def __str__(self):
        return str(self.to_entry())

    def __setitem__(self, key, item):
        self.data[key] = item

    def clear(self):
        self.data = {}

    def get(self, key, default=None):
        try:
            return self.data[key]
        except KeyError:
            return default

    def to_entry(self) -> Dict[str, Any]:
        """"Convert Page into serialised json for site.json"""

        def get(key):
            """return a dictionary item or None"""
            return self[key] if key in self else None

        items = ['source', 'target', 'page_type']
        entry = {}
        for item in items:
            if get(item) is not None:
                entry[item] = self[item]
        return entry



"""Stone Sites"""

import sys
from pathlib import Path

from stone.plugins import import_plugin

from stone.generators.blog import Generator
from stone.generators.single import Generator

_BUILTIN_GENERATORS = (
    'stone.generators.blog',
    'stone.generators.single',
)


def import_generators(module_name):
    if module_name in _BUILTIN_GENERATORS:
        return module_name

    filename = module_name + '.py'
    module_path = Path.home().joinpath('.stone', 'generators', filename)
    module_name = 'stone.ext.generators.' + module_name
    result = import_plugin(module_name, module_path)
    return result


def load_generator(name, *args, **kwargs):
    module_name = import_generators(name)
    return sys.modules[name].Generator(*args, **kwargs)


def get_module_name(name):
    builtin_name = 'stone.generators.' + name
    if builtin_name in _BUILTIN_GENERATORS:
        return builtin_name

    return 'stone.ext.generators.' + name



"""Site

Stone's representation of a website
"""
# -*- coding: utf-8 -*-

from collections import UserDict
from json import JSONEncoder
import os
from typing import Dict, List

from stone.page import Page, PageEncoder
from stone.resource import Resource


class SiteEncoder(JSONEncoder):
    """JSON encoder for Site"""

    def default(self, o):
        if isinstance(o, Site):
            items = ['site', 'type', 'source', 'target', 'templates']
            result = {}
            for item in items:
                try:
                    result[item] = o[item]
                except KeyError:
                    pass
            result['pages'] = o.pages
            return result
        if isinstance(o, Page):
            return PageEncoder().default(o)

        # When not a page, call the JSONEncoder. It will call the correct fail.
        return JSONEncoder.default(self, o)


class Site(UserDict):  # pylint: disable=too-many-ancestors
    """Representation of a Site"""

    root: str
    pages: List[Page]
    templates: List[str]
    resources: List[str]
    data: Dict[str, str]

    def __init__(self, root: str, data: Dict) -> None:
        super().__init__()
        self.root = root
        self.data = data
        self.pages = []
        self.templates = []
        self.resources = []
        self._parse(data)

    def __contains__(self, key):
        return str(key) in self.data

    def __repr__(self):
        return "Site(%r, %r)" % (self.root, self.data)

    def __str__(self):
        """Return the stringified version of Site

        Return a selective dictionary of Site as a string
        """
        items = ['site', 'type', 'source', 'target', 'templates']
        result = {}
        for item in items:
            try:
                result[item] = self[item]
            except KeyError:
                pass
        result['pages'] = self.pages
        return str(result)

    def __setitem__(self, key, item):
        self.data[str(key)] = item

    def _parse(self, data):
        """Load pages to be generated"""
        for page_data in data['pages']:
            self.pages.append(
                Page(
                    self,
                    page_data.pop('source'),
                    page_data.pop('target'),
                    data=page_data))

        self.templates = [
            os.path.join(self.root, template)
            for template in data.get('templates', [])
        ]
        self.data['templates'] = self.templates
        self.resources = [
            Resource(self.root, **resource)
            for resource in data.get('resources', [])
        ]
        self.data['resources'] = self.resources



"""Markdown to HTML Renderer"""

import mistune
import mistune_contrib.meta as m_meta


class Renderer:
    """Renders Markdown into HTML"""

    def __init__(self, _next=None, *args, **kwargs):
        self._next = _next

    def __eq__(self, other):
        return tuple(self) == tuple(self)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, *self)

    def __str__(self):
        return str(tuple(self))

    def render(self, page):
        """Renders a page content from Markdown to HTML"""
        meta, page['content'] = m_meta.parse(page['content'])
        page['content'] = mistune.markdown(page['content'])

        for key, value in meta.items():
            page[key] = value

        if self._next:
            return self._next.render(page)

        return page



"""Stone Blog"""

from stone.generators.abc import SiteGenerator


class Generator(SiteGenerator):
    """Stones interpretation of a blog"""
    site_type = "blog"

    def __init__(self):
        pass

    def __eq__(self, other):
        return tuple(self) == tuple(self)

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def __str__(self):
        return str(tuple(self))

    def generate(self, renderer, site):
        rendered_pages = []
        index = None

        for page in site.pages:
            if page.get('page_type') == 'index':
                index = page
                continue

            rendered_page = renderer.render(page)
            rendered_pages.append(rendered_page)

        index['posts'] = [post for post in site.pages if post is not index]
        index['posts'].reverse()
        rendered_pages.append(renderer.render(index))

        return rendered_pages



"""Stone Backends"""
import sys
from pathlib import Path

from stone.plugins import import_plugin
from stone.backends.file import Backend

_BUILTIN_BACKENDS = ('file')


def import_backend(module_name):
    if module_name in _BUILTIN_BACKENDS:
        return 'stone.backends.' + module_name

    filename = module_name + '.py'
    module_path = Path.home().joinpath('.stone', 'backends', filename)
    module_name = 'stone.ext.backends.' + module_name
    return import_plugin(module_name, module_path)


def load_backends(backends=[], *args, **kwargs):
    if not backends:
        backends = ['file']

    results = []
    for b in backends:
        module_name = import_backend(b)
        results.append(sys.modules[module_name].Backend(*args, **kwargs))

    return results



"""Stone Renderers"""
import sys
from pathlib import Path

from stone.plugins import import_plugin
from stone.renderers.jinja2 import Renderer as Jinja2Renderer
from stone.renderers.markdown import Renderer as MarkdownRenderer

_BUILTIN_RENDERERS = (
    'jinja2',
    'markdown',
)


def import_renderer(module_name):
    if module_name in _BUILTIN_RENDERERS:
        return 'stone.renderers.' + module_name

    filename = module_name + '.py'
    module_path = Path.home().joinpath('.stone', 'renderers', filename)
    module_name = 'stone.ext.renderers' + module_name
    return import_plugin(module_name, module_path)


def load_renderers(renderers=[], templates=[]):
    if not renderers:
        renderers = ['markdown', 'jinja2']

    previous_renderer = None
    for r in renderers[::-1]:
        module_name = import_renderer(r)
        renderer = sys.modules[module_name].Renderer(
            templates=templates, _next=previous_renderer)
        previous_renderer = renderer
    return previous_renderer



"""Stone Site Generator"""

from jinja2 import select_autoescape, Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound


class Renderer:
    """Renders Jinja2 templated input"""

    def __init__(self, _next=None, *args, **kwargs):
        self._next = _next
        try:
            self.templates = kwargs['templates']
        except KeyError:
            self.templates = []

    def __eq__(self, other):
        return tuple(self) == tuple(self)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, *self)

    def __str__(self):
        return str(tuple(self))

    def render(self, page):
        """Render using Jinja2 templates"""
        environment = Environment(
            loader=FileSystemLoader(self.templates),
            autoescape=select_autoescape(["html", "xml"]))

        try:
            page['content'] = environment.get_template(
                page['template']).render(page.data)
        except (TemplateNotFound, KeyError) as template_error:
            page['content'] = environment.from_string(page['content']).render(
                page.data)

        if self._next:
            return self._next.render(page)

        return page



"""Resource

Stone's representation for resources like CSS
"""

from collections import UserDict
import errno
import os


class Resource(UserDict):
    """Resource: Stones representation for resources like CSS"""

    def __init__(self, site_root, source, target, resource_type=None):
        super().__init__()
        self.data = {
            "resource_type": resource_type,
            "source": source,
            "source_path": os.path.abspath(os.path.join(site_root, source)),
            "target": target,
            "target_path": os.path.abspath(os.path.join(site_root, target)),
            "href": target.split('/')[1]
        }
        with open(self.data["source_path"], "r") as source_file:
            self.data["content"] = source_file.read()

    def __contains__(self, key):
        return str(key) in self.data

    def __missing__(self, key):
        if str(key) in self.data:
            return self.data
        raise KeyError(key)

    def __repr__(self):
        return "Resource(%r, %r)" % (self.data['source'], self.data['target'])

    def __setitem__(self, key, item):
        self.data[str(key)] = item

    def render(self):
        """Render resource to file"""
        print("Rendering: %s to %s" % (self.data['source_path'],
                                       self.data['target_path']))
        try:
            with open(self.data['target_path'], "w") as target_file:
                target_file.write(self.data["content"])
        except FileNotFoundError as fnf:
            if fnf.errno == errno.ENOENT:
                os.makedirs(
                    os.path.split(self.data['target_path'])[0], exist_ok=True)
                self.render()
            else:
                raise



"""Config

Stone's representation for site.json
"""

from collections import UserDict
import json
import os

from stone.site import SiteEncoder

CONFIG_VERSION = 1
SITE_CONFIG_FILE = "site.json"


class SiteConfig(UserDict):
    """SiteConfig

    A UserDict which requires a path
    """
    def __init__(self, path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path
        self.data = {k: v for k, v in kwargs.items()}

    def __repr__(self):
        class_name = type(self).__name__
        return "{}({!r})".format(class_name, self.path)

    def __missing__(self, key):
        return self[key]

    def __contains__(self, key):
        return key in self.data[key]

    def __setitem__(self, key, item):
        self.data[key] = item


def read(path):
    """Load site configuration"""
    configs = []
    json_data = json.loads(
        open(os.path.join(path, SITE_CONFIG_FILE), "r").read())

    try:
        if json_data["version"] != 1:
            raise ValueError('Version not supported. Supported version {}'.
                             format(CONFIG_VERSION))
        for site_data in json_data["sites"]:
            configs.append(SiteConfig(path, **site_data))
    except KeyError:
        configs.append(SiteConfig(path, **site_data))

    return configs


def write(path, sites):
    """Serialize a site to JSON"""
    config = {'version': CONFIG_VERSION}
    config['sites'] = sites

    with open(os.path.join(path, self.site_config_file), "w") as cfg_file:
        cfg_file.write(json.dumps(config, cls=SiteEncoder, indent=4))



"""Stone Static Content"""

from stone.generators.abc import SiteGenerator


class Generator(SiteGenerator):
    """Stones interpretation of a single page"""
    site_type = "single"

    def __init__(self):
        pass

    def __eq__(self, other):
        return tuple(self) == tuple(self)

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def __str__(self):
        return str(tuple(self))

    def generate(self, renderer, site):
        return [renderer.render(page) for page in site.pages]



"""Stone Plugin Functions"""

import importlib
import sys


def import_plugin(module_name, file_path):
    """Returns the name of the imported python module or raises ImportError"""
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)

    if module is None:
        raise ImportError("Could not find {}".format(module_name))

    spec.loader.exec_module(module)
    sys.modules[module_name] = module
    return module_name



"""Write strings to a file"""

import os
import errno


class Backend:
    """Stone backend for writing to file"""

    def __init__(self, *args, **kwargs):
        try:
            self._root = kwargs['root']
        except KeyError:
            self._root = None

    def __eq__(self, other):
        return tuple(self) == tuple(self)

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def __str__(self):
        return str(tuple(self))

    def _write(self, _path, content):
        path = os.path.join(self._root, _path)
        with open(path, 'w') as output:
            output.write(content)

    def commit(self, page):
        """Commit page to a file"""

        try:
            self._write(page['target_path'], page['content'])
        except FileNotFoundError as fnf:
            if fnf.errno == errno.ENOENT:
                os.makedirs(
                    os.path.split(page.data['target_path'])[0], exist_ok=True)
                self._write(page['target_path'], page['content'])
            else:
                raise
